/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.router;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Arrays;
import java.util.Collections;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelleTestFactory;
import de.ozgcloud.eingang.common.zufi.ZufiService;
import de.ozgcloud.eingang.router.errorhandling.AdapterConfigurationException;
import de.ozgcloud.eingang.router.errorhandling.UnknownOrganisationseinheitException;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;
import io.grpc.Channel;
import io.grpc.ManagedChannel;
import io.grpc.stub.AbstractStub;
import net.devh.boot.grpc.client.channelfactory.GrpcChannelFactory;
import net.devh.boot.grpc.client.inject.StubTransformer;
import net.devh.boot.grpc.client.stubfactory.StubFactory;

class VorgangManagerServerResolverTest {

	@Spy
	@InjectMocks
	private VorgangManagerServerResolver resolver;

	@Mock
	private GrpcChannelFactory channelFactory;
	@Mock
	private StubFactory stubFactory;

	@Mock
	private ZufiService zufiService;

	@Nested
	class TestFindStubFactory {

		@Mock
		private StubFactory stubFactory;

		@Test
		void shouldSetApplicableFactory() {
			when(stubFactory.isApplicable(any())).thenReturn(true);
			setStubFactories(stubFactory, stubFactory);

			resolver.findApplicableStubFactories();

			assertThat(ReflectionTestUtils.getField(resolver, "vorgangBlockingStubFactory")).isSameAs(stubFactory);
		}

		@Test
		void shouldThrowExceptionIfNotFound() {
			setStubFactories(stubFactory);

			assertThrows(AdapterConfigurationException.class, () -> resolver.findApplicableStubFactories());
		}

		private void setStubFactories(StubFactory... factories) {
			ReflectionTestUtils.setField(resolver, "stubFactories", Arrays.asList(factories));
		}
	}

	@DisplayName("Resolve vorgangManager service by organisationsEinheitenId")
	@Nested
	class TestResolveVorgangManagerServiceStubByOrganisationsEinheitenId {

		@Mock
		private ManagableStub<VorgangServiceBlockingStub> stub;

		private final Optional<String> organisationsEinheitenId = Optional.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);

		@BeforeEach
		void mock() {
			ReflectionTestUtils.setField(resolver, "vorgangBlockingStubFactory", stubFactory);

			doReturn(stub).when(resolver).createStub(any(), any(), any());
		}

		@Test
		void shouldCallCreateStub() {
			resolveStub();

			verify(resolver).createStub(organisationsEinheitenId, stubFactory, VorgangServiceBlockingStub.class);
		}

		@Test
		void shouldReturnStub() {
			var createdStub = resolveStub();

			assertThat(createdStub).isEqualTo(stub);
		}

		private ManagableStub<VorgangServiceBlockingStub> resolveStub() {
			return resolver.resolveVorgangServiceBlockingStubByOrganisationseinheitenId(organisationsEinheitenId);
		}
	}

	@DisplayName("Resolve binaryFile service by organisationsEinheitenId")
	@Nested
	class TestResolveBinaryFileServiceStubByOrganisationsEinheitenId {

		@Mock
		private ManagableStub<BinaryFileServiceStub> stub;

		private final Optional<String> organisationsEinheitenId = Optional.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);

		@BeforeEach
		void mock() {
			ReflectionTestUtils.setField(resolver, "binaryFileStubFactory", stubFactory);

			doReturn(stub).when(resolver).createStub(any(), any(), any());
		}

		@Test
		void shouldCallCreateStub() {
			resolveStub();

			verify(resolver).createStub(organisationsEinheitenId, stubFactory, BinaryFileServiceStub.class);
		}

		@Test
		void shouldReturnStub() {
			var createdStub = resolveStub();

			assertThat(createdStub).isEqualTo(stub);
		}

		private ManagableStub<BinaryFileServiceStub> resolveStub() {
			return resolver.resolveBinaryFileServiceStubByOrganisationsEinheitId(organisationsEinheitenId);
		}
	}

	@DisplayName("Create stub")
	@Nested
	class TestCreateStub {

		@Mock
		private ManagableStub<?> stub;

		private final Optional<String> organisationsEinheitenId = Optional.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);

		@DisplayName("on zufi strategy")
		@Nested
		class TestOnZufiRoutingStrategy {

			@BeforeEach
			void mock() {
				doReturn(stub).when(resolver).createCloseableStub(any(), any(), any());
				setProperties(VorgangManagerListPropertiesTestFactory.createForZufiRouting());
			}

			@Test
			void shouldCallCreateStub() {
				createStub();

				verify(resolver).createStub(organisationsEinheitenId, stubFactory, VorgangServiceBlockingStub.class);
			}

			@Test
			void shouldReturnStub() {
				var createdStub = createStub();

				assertThat(createdStub).isEqualTo(stub);
			}
		}

		@DisplayName("On other routing strategy")
		@Nested
		class TestOnOtherRoutingStrategy {

			@BeforeEach
			void mock() {
				doReturn(stub).when(resolver).createStubByConfiguredChannels(any(), any(), any());
				setProperties(VorgangManagerListPropertiesTestFactory.createForSingleRouting());
			}

			@Test
			void shouldCallCreateStub() {
				createStub();

				verify(resolver).createStubByConfiguredChannels(organisationsEinheitenId, stubFactory, VorgangServiceBlockingStub.class);
			}

			@Test
			void shouldReturnStub() {
				var createdStub = createStub();

				assertThat(createdStub).isEqualTo(stub);
			}
		}

		private ManagableStub<VorgangServiceBlockingStub> createStub() {
			return resolver.createStub(organisationsEinheitenId, stubFactory, VorgangServiceBlockingStub.class);
		}
	}

	@DisplayName("Create closeable stub")
	@Nested
	class TestCreateCloseableStub {

		@Mock
		private ManagableStub<?> manageableStub;
		@Mock
		private AbstractStub<?> stub;
		@Mock
		private ManagedChannel managedChannel;

		private Optional<String> organisationsEinheitenId = Optional.of(VorgangManagerListPropertiesTestFactory.ORGANISATIONSEINHEIT_ID);
		private final String vorgangManagerAddress = "dummyVorgangManagerAddress";

		@BeforeEach
		void mock() {
			doReturn(vorgangManagerAddress).when(resolver).getVorgangManagerAddress(any());
			doReturn(managedChannel).when(resolver).createChannel(any());
			doReturn(stub).when(resolver).applyStubTransformers(any(), any());
		}

		@Test
		void shouldGetVorgangManagerAddress() {
			createCloseableStub();

			verify(resolver).getVorgangManagerAddress(organisationsEinheitenId);
		}

		@Test
		void shouldCreateChannel() {
			createCloseableStub();

			verify(resolver).createChannel(vorgangManagerAddress);
		}

		@Test
		void shouldCreateStub() {
			createCloseableStub();

			verify(stubFactory).createStub(VorgangServiceBlockingStub.class, managedChannel);
		}

		@Test
		void shouldApplStubTransformers() {
			createCloseableStub();

			verify(resolver).applyStubTransformers(any(), eq(vorgangManagerAddress));
		}

		@Test
		void shouldReturnStub() {
			var createdStub = (ClosableStub) createCloseableStub();

			assertThat(createdStub).isNotNull();
			assertThat(createdStub.get()).isEqualTo(stub);
		}

		private ManagableStub<VorgangServiceBlockingStub> createCloseableStub() {
			return resolver.createCloseableStub(organisationsEinheitenId, stubFactory, VorgangServiceBlockingStub.class);
		}
	}

	@DisplayName("Get vorgangManager address")
	@Nested
	class TestGetVorgangManagerAddress {

		private final Optional<String> organisationsEinheitenId = Optional.of(VorgangManagerListPropertiesTestFactory.ORGANISATIONSEINHEIT_ID);
		private final String vorgangManagerAddress = "DummyVorgangManagerAddress";

		@Test
		void shouldCallZufiService() {
			when(zufiService.getVorgangManagerUrl(any())).thenReturn(vorgangManagerAddress);

			getVorgangManagerAddress();

			verify(zufiService).getVorgangManagerUrl(VorgangManagerListPropertiesTestFactory.ORGANISATIONSEINHEIT_ID);
		}

		@Test
		void shouldThrowExceptionIfOrganisationsEinheitIsNotPresent() {
			var emptyOrganisationsEinheitId = Optional.<String>empty();

			assertThatThrownBy(() -> resolver.getVorgangManagerAddress(emptyOrganisationsEinheitId)).isInstanceOf(TechnicalException.class);
		}

		@Test
		void shouldReturnAddress() {
			when(zufiService.getVorgangManagerUrl(any())).thenReturn(vorgangManagerAddress);

			var address = getVorgangManagerAddress();

			assertThat(address).isEqualTo(vorgangManagerAddress);

		}

		private String getVorgangManagerAddress() {
			return resolver.getVorgangManagerAddress(organisationsEinheitenId);
		}
	}

	@DisplayName("Create stub by configured channels")
	@Nested
	class TestCreateStubByConfiguredChannels {

		@Mock
		private Channel channel;
		@Mock
		private StubFactory stubFactory;
		@Mock
		private AbstractStub<?> createdStub;

		private Class<? extends AbstractStub<?>> stubClass = VorgangServiceBlockingStub.class;

		@BeforeEach
		void initTest() {
			doReturn(VorgangManagerListPropertiesTestFactory.CHANNEL_NAME).when(resolver).getChannelName(any());
			doReturn(channel).when(resolver).createChannelByName(any());
			setProperties(VorgangManagerListPropertiesTestFactory.createForSingleRouting());
			doReturn(createdStub).when(resolver).applyStubTransformers(any(), any());
		}

		@Test
		void shouldGetChannel() {
			createStubByConfiguredChannels();

			verify(resolver).createChannelByName(VorgangManagerListPropertiesTestFactory.CHANNEL_NAME);
		}

		@Test
		void shouldApplyTransformers() {
			createStubByConfiguredChannels();

			verify(resolver).applyStubTransformers(any(), any());
		}

		@Test
		void shouldCreateStubByFactory() {
			createStubByConfiguredChannels();

			verify(stubFactory).createStub(eq(stubClass), any());
		}

		@Test
		void shouldReturnStub() {
			var stub = createStubByConfiguredChannels();

			assertThat(stub.get()).isEqualTo(createdStub);
		}

		private ManagableStub<?> createStubByConfiguredChannels() {
			return resolver.createStubByConfiguredChannels(Optional.of(VorgangManagerListPropertiesTestFactory.ORGANISATIONSEINHEIT_ID), stubFactory,
					VorgangServiceBlockingStub.class);
		}
	}

	@Nested
	class TestGetChannelName {

		@Test
		void shouldUseSingleName() {
			setProperties(VorgangManagerListPropertiesTestFactory.createForSingleRouting());

			var name = resolver.getChannelName(Optional.empty());

			assertThat(name).contains(VorgangManagerListPropertiesTestFactory.CHANNEL_NAME);
		}

		@Test
		void shouldUseNameFromMap() {
			setProperties(VorgangManagerListPropertiesTestFactory.createForMultiRouting());

			var name = resolver.getChannelName(Optional.of(VorgangManagerListPropertiesTestFactory.ORGANISATIONSEINHEIT_ID));

			assertThat(name).contains(VorgangManagerListPropertiesTestFactory.CHANNEL_NAME);
		}

		@Test
		void shouldGetFundstellenName() {
			setProperties(VorgangManagerListPropertiesTestFactory.createWithFundstelle());

			var name = resolver.getChannelName(Optional.of("4711"));

			verify(resolver).getFundstelleChannelName();
			assertThat(name).isEqualTo(VorgangManagerListPropertiesTestFactory.FUNDSTELLE_CHANNEL_NAME);
		}
	}

	@Nested
	class TestGetFundstellenChannelName {

		@Nested
		class TestStrategyFundstelle {

			@Test
			void shouldCreateChannel() {
				setProperties(VorgangManagerListPropertiesTestFactory.createWithFundstelle());

				var channel = resolver.getFundstelleChannelName();

				assertThat(channel).isNotNull().isEqualTo(VorgangManagerListPropertiesTestFactory.FUNDSTELLE_CHANNEL_NAME);
			}

			@Test
			void shouldThrowExceptionIfFundstelleIsMissing() {
				var props = VorgangManagerListPropertiesTestFactory.createWithFundstelle();
				props.setFundstelleVorgangManagerName(Optional.empty());
				setProperties(props);

				assertThrows(AdapterConfigurationException.class, () -> resolver.getFundstelleChannelName());
			}
		}

		@Nested
		class TestStrategyDeny {

			@Test
			void shouldThrowException() {
				setProperties(VorgangManagerListPropertiesTestFactory.createForMultiRouting());

				assertThrows(UnknownOrganisationseinheitException.class, () -> resolver.getFundstelleChannelName());
			}
		}
	}

	@Nested
	class TestCreateChannelByName {

		@Test
		void shouldCallChannelFactory() {
			resolver.createChannelByName(VorgangManagerListPropertiesTestFactory.VORGANG_MANAGER_NAME);

			verify(channelFactory).createChannel(VorgangManagerListPropertiesTestFactory.VORGANG_MANAGER_NAME);
		}
	}

	@Nested
	class TestApplyStubTransformer {

		@Mock
		private AbstractStub<?> stub;
		@Mock
		private StubTransformer transformer;

		@BeforeEach
		void init() {
			ReflectionTestUtils.setField(resolver, "stubTransformers", Collections.singleton(transformer));
		}

		@Test
		void shouldCallTransform() {
			resolver.applyStubTransformers(stub, VorgangManagerListPropertiesTestFactory.CHANNEL_NAME);

			verify(transformer).transform(VorgangManagerListPropertiesTestFactory.CHANNEL_NAME, stub);
		}
	}

	private void setProperties(VorgangManagerListProperties properties) {
		ReflectionTestUtils.setField(resolver, "properties", properties);
	}
}